use crate::prelude::*;
use anyhow::Result;
use ecs::{Resource, resources::Resource as ResourceTrait};
use std::collections::HashMap;

#[derive(Debug)]
pub enum EngineEvent<'a> {
    MouseMotion(u32, u32),
    MouseButtonDown(MouseButton),
    MouseButtonUp(MouseButton),
    MouseWheel(i32, i32, MouseWheelDirection),

    KeyDown(Keycode),
    KeyUp(Keycode),

    ControllerButtonDown(&'a Controller, ControllerButton),
    ControllerButtonUp(&'a Controller, ControllerButton),
    ControllerAxis(&'a Controller, ControllerAxis),
    ControllerAdded(&'a Controller),
    ControllerRemoved(&'a Controller),

    JoystickButtonDown(&'a Joystick),
    JoystickButtonUp(&'a Joystick),
    JoystickAxis(&'a Joystick, u8, i16),
    JoystickAdded(&'a Joystick),
    JoystickRemoved(&'a Joystick),

    FileDrop(String),
}

#[derive(Resource)]
pub struct InputMap {
    pub direction_mapping: HashMap<Keycode, GuiDirection>,
}

pub trait EventSystem<Func, T> {
    fn event_system(world_builder: &mut WorldBuilder, func: Func);
}

macro_rules! impl_event_system {
    ( $( $t:ident, )* ) => {
        paste::paste! {
            impl<Func, T, $( $t, )* > EventSystem<Func, (T, $( $t, )*)> for Engine
            where
                Func: Fn(&mut T, &mut Commands, EngineEvent<'_>, $( &mut $t, )* ) -> Result<()> + Send + Sync + 'static,
                T: ResourceTrait,
                $(
                    $t: ResourceTrait,
                )*
            {
                fn event_system(world_builder: &mut WorldBuilder, func: Func) {
                    world_builder.add_system(
                        0,
                        move |world: &World,
                        commands: &mut Commands,
                        consumer: &mut T,
                        gui_handler: &mut GuiHandler,
                        input_map: &mut InputMap,
                        context: &mut Context,
                        $(
                            [< $t:lower >]: &mut $t,
                        )*
                        |
                        -> Result<bool> {
                            let res = context.events(world, |event| {
                                match event {
                                    Event::MouseMotion(x, y) => {
                                        gui_handler.set_mouse_pos(x, y)?;
                                        func(
                                            consumer,
                                            commands,
                                            EngineEvent::MouseMotion(x, y),
                                            $(
                                                [< $t:lower >],
                                            )*
                                        )?;
                                    }
                                    Event::MouseButtonDown(mouse_button) => {
                                        if !gui_handler.mouse_down(mouse_button)? {
                                            func(
                                                consumer,
                                                commands,
                                                EngineEvent::MouseButtonDown(mouse_button),
                                                $(
                                                    [< $t:lower >],
                                                )*
                                            )?;
                                        }
                                    }
                                    Event::MouseButtonUp(mouse_button) => {
                                        if !gui_handler.mouse_up(mouse_button)? {
                                            func(
                                                consumer,
                                                commands,
                                                EngineEvent::MouseButtonUp(mouse_button),
                                                $(
                                                    [< $t:lower >],
                                                )*
                                            )?;
                                        }
                                    }
                                    Event::MouseWheel(x, y, direction) => {
                                        func(
                                            consumer,
                                            commands,
                                            EngineEvent::MouseWheel(x, y, direction),
                                            $(
                                                [< $t:lower >],
                                            )*
                                        )?;
                                    }
                                    Event::KeyDown(keycode) => {
                                        if let Some(direction) =
                                            input_map.direction_mapping.get(&keycode)
                                        {
                                            if gui_handler.update_selection(*direction)? {
                                                return Ok(());
                                            }
                                        }

                                        match keycode {
                                            Keycode::Backspace => {
                                                if gui_handler.remove_char()? {
                                                    return Ok(());
                                                }
                                            }
                                            Keycode::Return => {
                                                if gui_handler.accept_selection()? {
                                                    return Ok(());
                                                }
                                            }
                                            Keycode::Escape => {
                                                if gui_handler.decline_topgui(commands)? {
                                                    return Ok(());
                                                }
                                            }

                                            _ => (),
                                        }

                                        func(
                                            consumer,
                                            commands,
                                            EngineEvent::KeyDown(keycode),
                                            $(
                                                [< $t:lower >],
                                            )*
                                        )?;
                                    }
                                    Event::KeyUp(keycode) => {
                                        if input_map.direction_mapping.get(&keycode).is_some()
                                            && gui_handler.update_selection(GuiDirection::None)?
                                        {
                                            return Ok(());
                                        }

                                        func(
                                            consumer,
                                            commands,
                                            EngineEvent::KeyUp(keycode),
                                            $(
                                                [< $t:lower >],
                                            )*
                                        )?;
                                    }
                                    Event::TextInput(text) => {
                                        if let Some(writeable) = gui_handler.writeable() {
                                            for c in text.chars() {
                                                writeable.add_letter(gui_handler, c)?;
                                            }
                                        }
                                    }
                                    Event::ControllerButtonDown(controller, button) => {
                                        if gui_handler.check_navigatable() {
                                            match button {
                                                ControllerButton::A => {
                                                    if gui_handler.accept_selection()? {
                                                        return Ok(());
                                                    }
                                                }

                                                ControllerButton::B => {
                                                    if gui_handler.decline_topgui(commands)? {
                                                        return Ok(());
                                                    }
                                                }

                                                ControllerButton::RightButton => {
                                                    if gui_handler.next_tab_topgui(commands, false)? {
                                                        return Ok(());
                                                    }
                                                }

                                                ControllerButton::LeftButton => {
                                                    if gui_handler
                                                        .previous_tab_topgui(commands, false)?
                                                    {
                                                        return Ok(());
                                                    }
                                                }

                                                ControllerButton::RightTrigger => {
                                                    if gui_handler.next_tab_topgui(commands, true)? {
                                                        return Ok(());
                                                    }
                                                }

                                                ControllerButton::LeftTrigger => {
                                                    if gui_handler
                                                        .previous_tab_topgui(commands, true)?
                                                    {
                                                        return Ok(());
                                                    }
                                                }

                                                ControllerButton::DPadDown => {
                                                    let selection_res = gui_handler
                                                        .update_selection(GuiDirection::Down)?;
                                                    gui_handler.update_selection(GuiDirection::None)?;

                                                    if selection_res {
                                                        return Ok(());
                                                    }
                                                }

                                                ControllerButton::DPadUp => {
                                                    let selection_res = gui_handler
                                                        .update_selection(GuiDirection::Up)?;
                                                    gui_handler.update_selection(GuiDirection::None)?;

                                                    if selection_res {
                                                        return Ok(());
                                                    }
                                                }

                                                ControllerButton::DPadRight => {
                                                    let selection_res = gui_handler
                                                        .update_selection(GuiDirection::Right)?;
                                                    gui_handler.update_selection(GuiDirection::None)?;

                                                    if selection_res {
                                                        return Ok(());
                                                    }
                                                }

                                                ControllerButton::DPadLeft => {
                                                    let selection_res = gui_handler
                                                        .update_selection(GuiDirection::Left)?;
                                                    gui_handler.update_selection(GuiDirection::None)?;

                                                    if selection_res {
                                                        return Ok(());
                                                    }
                                                }

                                                _ => (),
                                            }

                                            if !gui_handler.accept_custom_selection(commands, button)? {
                                                func(
                                                    consumer,
                                                    commands,
                                                    EngineEvent::ControllerButtonDown(
                                                        controller, button,
                                                    ),
                                                    $(
                                                        [< $t:lower >],
                                                    )*
                                                )?;
                                            }
                                        } else {
                                            func(
                                                consumer,
                                                commands,
                                                EngineEvent::ControllerButtonDown(controller, button),
                                                $(
                                                    [< $t:lower >],
                                                )*
                                            )?;
                                        }
                                    }
                                    Event::ControllerButtonUp(controller, button) => {
                                        func(
                                            consumer,
                                            commands,
                                            EngineEvent::ControllerButtonUp(controller, button),
                                            $(
                                                [< $t:lower >],
                                            )*
                                        )?;
                                    }
                                    Event::ControllerAxis(controller) => {
                                        if !gui_handler.check_navigatable() {
                                            func(
                                                consumer,
                                                commands,
                                                EngineEvent::ControllerAxis(
                                                    controller,
                                                    controller.controller_axis(),
                                                ),
                                                $(
                                                    [< $t:lower >],
                                                )*
                                            )?;
                                        } else {
                                            gui_handler.update_selection(controller.direction())?;
                                        }
                                    }
                                    Event::ControllerAdded(controller) => {
                                        func(
                                            consumer,
                                            commands,
                                            EngineEvent::ControllerAdded(controller),
                                            $(
                                                [< $t:lower >],
                                            )*
                                        )?;
                                    }
                                    Event::ControllerRemoved(controller) => {
                                        func(
                                            consumer,
                                            commands,
                                            EngineEvent::ControllerRemoved(controller),
                                            $(
                                                [< $t:lower >],
                                            )*
                                        )?;
                                    }

                                    Event::JoystickAxis(joystick, axis_index, value) => {
                                        func(
                                            consumer,
                                            commands,
                                            EngineEvent::JoystickAxis(joystick, axis_index, value),
                                            $(
                                                [< $t:lower >],
                                            )*
                                        )?;
                                    }
                                    Event::JoystickButtonDown(joystick) => {
                                        func(
                                            consumer,
                                            commands,
                                            EngineEvent::JoystickButtonDown(joystick),
                                            $(
                                                [< $t:lower >],
                                            )*
                                        )?;
                                    }
                                    Event::JoystickButtonUp(joystick) => {
                                        func(
                                            consumer,
                                            commands,
                                            EngineEvent::JoystickButtonUp(joystick),
                                            $(
                                                [< $t:lower >],
                                            )*
                                        )?;
                                    }
                                    Event::JoystickAdded(joystick) => {
                                        func(
                                            consumer,
                                            commands,
                                            EngineEvent::JoystickAdded(joystick),
                                            $(
                                                [< $t:lower >],
                                            )*
                                        )?;
                                    }
                                    Event::JoystickRemoved(joystick) => {
                                        func(
                                            consumer,
                                            commands,
                                            EngineEvent::JoystickRemoved(joystick),
                                            $(
                                                [< $t:lower >],
                                            )*
                                        )?;
                                    }

                                    Event::FileDrop(filename) => {
                                        func(
                                            consumer,
                                            commands,
                                            EngineEvent::FileDrop(filename),
                                            $(
                                                [< $t:lower >],
                                            )*
                                        )?;
                                    }
                                }

                                Ok(())
                            })?;

                            gui_handler.process_callbacks(commands)?;

                            Ok(res)
                        },
                    )
                }
            }
        }
    };
}

impl_event_system!();
impl_event_system!(A, B,);
impl_event_system!(A, B, C,);
impl_event_system!(A, B, C, D,);
impl_event_system!(A, B, C, D, E,);
impl_event_system!(A, B, C, D, E, F,);
