use anyhow::Result;

use ecs::*;
use engine::prelude::*;

fn main() -> Result<()> {
    let mut world_builder = World::builder();

    Engine::new(EngineCreateInfo::default(), &mut world_builder)?;
    Engine::event_system(&mut world_builder, GameState::event);

    world_builder.add_system(10, GameState::update);
    world_builder.resources.insert(GameState::default());

    let scene: &mut Scene = world_builder.resources.get_mut()?;
    let view = scene.view_mut();
    let camera_control = FreeCameraControl::new(view)?;
    world_builder.resources.insert(camera_control);

    world_builder.build().run()
}

#[derive(Default, Resource)]
enum GameState {
    #[default]
    Startup,

    Game(Game),
}

impl GameState {
    fn update(commands: &mut Commands, game_state: &mut GameState) -> Result<bool> {
        match game_state {
            GameState::Startup => *game_state = GameState::Game(Game),
            GameState::Game(game) => game.update(commands)?,
        }

        Ok(true)
    }

    fn event(
        &mut self,
        commands: &mut Commands,
        event: EngineEvent<'_>,
        camera_control: &mut FreeCameraControl,
        scene: &mut Scene,
    ) -> Result<()> {
        match self {
            GameState::Startup => (),
            GameState::Game(game) => game.event(commands, event, camera_control, scene)?,
        }

        Ok(())
    }
}

struct Game;

impl Game {
    fn update(&mut self, _commands: &mut Commands) -> Result<()> {
        Ok(())
    }

    fn event(
        &mut self,
        _commands: &mut Commands,
        event: EngineEvent<'_>,
        camera_control: &mut FreeCameraControl,
        scene: &mut Scene,
    ) -> Result<()> {
        match event {
            EngineEvent::MouseButtonDown(MouseButton::Left) => {
                camera_control.mouse_down();
            }
            EngineEvent::MouseButtonUp(MouseButton::Left) => {
                camera_control.mouse_release();
            }
            EngineEvent::MouseMotion(x, y) => {
                camera_control.mouse_move(x, y, scene.view_mut())?;
            }

            _ => (),
        }

        Ok(())
    }
}
